#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''
__all__ = [ 'settings' ]

import decouple

class Settings(object):
    auto_cast = True

    def __getattr__(self, item):
        if item in dir(self):
            return super().__getattr__(item)
        else:
            return self._retrieve_option_from_config(item)

    def _retrieve_option_from_config(self, *args, **kwargs):
        option = decouple.config(*args, **kwargs)
        if self.auto_cast and \
                isinstance(option, str) and \
                option.upper() in ( 'TRUE', 'FALSE', ):
            return option.upper() == 'TRUE'
        else:
            return option

    def __call__(self, *args, **kwargs):
        if kwargs.get('force_reload', False):
            self.reload()
        return self._retrieve_option_from_config(*args, **kwargs)

    def reload(self):
        decouple.config.config = None

settings = Settings()
del Settings



#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''
from .facade import Facade
from .observer import Notifier

class Proxy(Notifier):
    NAME = None
    def __init__(self, name=None, data=None, *args, **kwargs):
        self._facade = Facade.get_instance()
        self._name = name or self.NAME
        self._data = data
        assert self._name is not None, f'Proxy name cannot be "None"'
        assert self._data is not None, f'Proxy data cannot be "None"'

    @property
    def facade(self):
        return self._facade

    @property
    def data(self):
        return self._data

    def on_register(self, *args, **kwargs):
        pass

    def on_remove(self, *args, **kwargs):
        pass



#!/usr/bin/env python3
# coding: utf-8
__all__ = [
    'timed_cache',
    'generate_password',
    'generate_salt',
    'AttrDict',
]

import functools
import os
import base64
import json
import string
import random
from datetime import datetime, timedelta

def timed_cache(maxsize=128, typed=False, **timedelta_kwargs):
    '''
    Usage:
    @timed_cache(hours=1)
    def get_something():
        pass
    '''
    def _wrapper(f):
        update_delta = timedelta(**timedelta_kwargs)
        next_update = datetime.utcnow() + update_delta
        f = functools.lru_cache(maxsize=maxsize, typed=typed)(f)

        @functools.wraps(f)
        def _inner_wrapper(*args, **kwargs):
            nonlocal update_delta, next_update
            now = datetime.utcnow()
            if now >= next_update:
                f.cache_clear()
                next_update = now + update_delta
            return f(*args, **kwargs)
        return _innter_wrapper
    return _wrapper

def generate_password(length=64, candidates=[]):
    candidates = candidates if candidates \
            else string.ascii_letters + string.digits + '!@#$%&'
    return ''.join(random.choices(candidates, k=length))

def generate_salt():
    return base64.b64encode(os.urandom(32)).decode()


class AttrDict(object):
    def __init__(self, obj):
        assert isinstance(obj, dict), f'Object to be wrapped should be a `dict` type'
        self._origin = obj

    def __contains__(self, item):
        return item in self.__dict__ \
                or item in self._origin

    def __getattr__(self, name):
        if name in self.__dict__ or name.startswith('_'):
            return self.__dict__[name]
        elif hasattr(self._origin, name):
            return getattr(self._origin, name)
        else:
            try:
                obj = self._origin[name]
                return AttrDict(obj) if isinstance(obj, dict) else obj
            except KeyError as e:
                raise AttributeError(f'Attribute \'{ name }\' is not found in ' \
                        f'class \'{ self.__class__.__name__ }\'.')

    def __setattr__(self, name, value):
        if name in self.__dict__ or name.startswith('_'):
            self.__dict__.update({ name: value })
        else:
            self._origin[name] = value

    def __delattr__(self, name):
        if name in self.__dict__ or name.startswith('_'):
            del self.__dict__[name]
        else:
            del self._origin[name]

    def __getitem__(self, name):
        obj = self._origin[name]
        return AttrDict(obj) if isinstance(obj, dict) else obj

    def __setitem__(self, name, value):
        self._origin[name] = value._origin if isinstance(value, AttrDict) else value

    def __delitem__(self, name):
        del self._origin[name]

    def __repr__(self):
        return json.dumps(self._origin, indent=2)




#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''
__all__ = [
    'session',
    'requests_retry_session',
    'restore_session',
    'persist_session',
    'UniRequest',
]

import requests
import json
import base64
import pickle
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry
from unipark.config import settings

def requests_retry_session(
    retries=None,
    backoff_factor=0.3,
    status_forcelist=(500, 502, 504),
    session=None,
    ):
    session = session or requests.Session()
    retries = int(settings('REQUEST_RETRIES', default=0)) \
            if retries == None else retries 
    retry = Retry(
        total=retries,
        read=retries,
        connect=retries,
        backoff_factor=backoff_factor,
        status_forcelist=status_forcelist,
        )
    adapter = HTTPAdapter(max_retries=retry)
    session.mount('http://', adapter)
    session.mount('https://', adapter)
    return session

def persist_session(session, filepath):
    data_obj = {
        'headers': dict(session.headers) \
                if hasattr(session, 'headers') else None,
        'cookies': base64.b64encode(pickle.dumps(session.cookies)).decode() \
                if session.cookies.get_dict() else None,
        'auth': getattr(session, 'auth', None),
        'timeout': getattr(session, 'timeout', None),
        'proxies': getattr(session, 'proxies', None),
        'cert': getattr(session, 'cert', None),
    }
    with open(filepath, 'wb') as fp:
        fp.write(json.dumps(data_obj).encode())

def restore_session(filepath):
    with open(filepath, 'rb') as fp:
        session_obj = json.loads(fp.read())
        session = requests.Session()

        headers = session_obj.get('headers', None)
        if headers:
            session.headers.update(headers)

        cookies = session_obj.get('cookies', None)
        if cookies:
            session.cookies.update(pickle.loads(
                    base64.b64decode(cookies)))

        auth = session_obj.get('auth', None)
        if auth:
            session.auth = auth

        timeout = session_obj.get('timeout', \
                settings('DEFAULT_TIMEOUT', default=300))
        if timeout:
            session.timeout = timeout

        proxies = session_obj.get('proxies', None)
        if proxies:
            session.proxies = proxies

        cert = session_obj.get('cert', None)
        if cert:
            session.cert = cert
    return requests_retry_session(session=session)

session = requests_retry_session()


class UniRequest(object):
    def __init__(self, *args, **kwargs):
        pass

    @staticmethod
    def head(url, session=None, *args, **kwargs):
        '''
        Same as GET, but transfers the status line and header section only.
        '''
        return requests_retry_session(session=session).head(url, *args, **kwargs)

    @staticmethod
    def options(url, session=None, *args, **kwargs):
        '''
        Describes the communication options for the target resource.
        '''
        return requests_retry_session(session=session).options(url, *args, **kwargs)

    @staticmethod
    def get(url, session=None, *args, **kwargs):
        '''
        Retrieve information from the given URL.
        '''
        return requests_retry_session(session=session).get(url, *args, **kwargs)

    @staticmethod
    def post(url, session=None, *args, **kwargs):
        '''
        Send data to server.
        '''
        return requests_retry_session(session=session).post(url, *args, **kwargs)

    @staticmethod
    def put(url, session=None, *args, **kwargs):
        '''
        Replaces all current representations of the target resource 
        with the uploaded content.
        '''
        return requests_retry_session(session=session).put(url, *args, **kwargs)

    @staticmethod
    def delete(url, session=None, *args, **kwargs):
        '''
        Removes all current representations of the target resource 
        given by a URI.
        '''
        return requests_retry_session(session=session).delete(url, *args, **kwargs)

    @staticmethod
    def patch(url, session=None, *args, **kwargs):
        '''
        Apply partial modifications to a resource
        '''
        return requests_retry_session(session=session).patch(url, *args, **kwargs)



#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''
import os
import unittest
from unipark.config import settings

def init_dotenv():
    with open('.env', 'wb') as fp:
        fp.write(
        b'# Generate by test_config.py\n'
        b'DEBUG=True\n'
        b'VAR_INTEGER=8\n'
        b'VAR_FLOAT=8.88\n'
        b'VAR_STRING=string from ".env"\n')

def override_dotenv():
    with open('.env', 'wb') as fp:
        fp.write(
        b'# Generate by test_config.py\n'
        b'DEBUG=False\n'
        b'VAR_INTEGER=6\n'
        b'VAR_FLOAT=6.66\n'
        b'VAR_STRING=string from ".env(override)"\n')

def deinit_dotenv():
    with open('.env', 'wb') as fp:
        fp.write(b'')

def init_envvar():
    os.environ.update({
        'DEBUG': 'False',
        'VAR_INTEGER': '6',
        'VAR_FLOAT': '6.66',
        'VAR_STRING': 'string from "os.environ"',
        })

def deinit_envvar():
    os.environ.pop('DEBUG', None)
    os.environ.pop('VAR_INTEGER', None)
    os.environ.pop('VAR_FLOAT', None)
    os.environ.pop('VAR_STRING', None)


class ConfigTest(unittest.TestCase):
    def setUp(self):
        deinit_dotenv()
        deinit_envvar()
        settings.reload()

    def tearDown(self):
        pass

    def test_001_retrieve_option_by_attribute(self):
        init_dotenv()

        self.assertEqual(settings.DEBUG, True)
        self.assertEqual(settings.VAR_INTEGER, '8')
        self.assertEqual(settings.VAR_FLOAT, '8.88')
        self.assertEqual(settings.VAR_STRING, 'string from ".env"')

    def test_002_retrieve_option_by_get(self):
        init_dotenv()

        self.assertEqual(settings('DEBUG', cast=bool), True)
        self.assertEqual(settings('VAR_INTEGER', cast=int), 8)
        self.assertEqual(settings('VAR_FLOAT', cast=float), 8.88)
        self.assertEqual(settings('VAR_STRING'), 'string from ".env"')

    def test_003_retrieve_option_priority(self):
        init_dotenv()
        init_envvar()

        self.assertEqual(settings('DEBUG', cast=bool), False)
        self.assertEqual(settings('VAR_INTEGER', cast=int), 6)
        self.assertEqual(settings('VAR_FLOAT', cast=float), 6.66)
        self.assertEqual(settings('VAR_STRING'), 'string from "os.environ"')

    def test_004_reload_dotenv(self):
        init_dotenv()
        self.assertEqual(settings.DEBUG, True)
        self.assertEqual(settings.VAR_INTEGER, '8')
        self.assertEqual(settings.VAR_FLOAT, '8.88')
        self.assertEqual(settings.VAR_STRING, 'string from ".env"')

        override_dotenv()
        settings.reload()
        self.assertEqual(settings.DEBUG, False)
        self.assertEqual(settings.VAR_INTEGER, '6')
        self.assertEqual(settings.VAR_FLOAT, '6.66')
        self.assertEqual(settings.VAR_STRING, 'string from ".env(override)"')



#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''
import unittest
from unipark.requests import (
    session, restore_session, persist_session, UniRequest
)

class RequestTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_001_persist_and_restore_session(self):
        pass

    def test_002_http_request(self):
        resp = UniRequest.head('http://captive.apple.com')
        self.assertEqual(resp.status_code, 200)

        resp = UniRequest.get('http://www.msftncsi.com/ncsi.txt')
        self.assertEqual(resp.status_code, 200)



#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''
import unittest
from unipark.utils import (
    generate_password, generate_salt, AttrDict
)

class UtilsTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_001_generator(self):
        password = generate_password()
        self.assertEqual(len(password), 64)

        password = generate_password(length=32)
        self.assertEqual(len(password), 32)

        salt = generate_salt()
        self.assertEqual(len(salt), 44)

    def test_002_attr_dict(self):
        MESSAGE = 'This is a string'
        inner = {
            'message': MESSAGE,
        }

        # 
        ad = AttrDict(inner)
        self.assertTrue(hasattr(ad, 'message'))
        self.assertIn('message', ad)
        self.assertEqual(ad.message, MESSAGE)
        self.assertEqual(ad['message'], MESSAGE)
        self.assertEqual(ad.get('message'), MESSAGE)

        del ad['message']
        self.assertFalse(hasattr(ad, 'message'))
        self.assertNotIn('message', ad)



#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''
import unittest
from unipark.mvc import Model, Controller, View
from unipark.mvc.command import SimpleCommand, MacroCommand
from unipark.mvc.proxy import Proxy
from unipark.mvc.mediator import Mediator
from unipark.mvc.facade import Facade

class TestObject(object):
    def __init__(self, value):
        self.input = value
        self.output = None


class TestSimpleCommand(SimpleCommand):
    NAME = 'CMD_TEST_SIMPLE'

    def execute(self, note):
        obj = note.body.get('obj')
        increase = note.body.get('increase')
        obj.output = obj.input + increase


class TestSubCommand1(SimpleCommand):
    def execute(self, note):
        obj = note.body.get('obj')
        increase = note.body.get('increase')
        if obj.output is None:
            obj.output = dict()
        obj.output['result1'] = obj.input + increase


class TestSubCommand2(SimpleCommand):
    def execute(self, note):
        obj = note.body.get('obj')
        multiple = note.body.get('multiple')
        if obj.output is None:
            obj.output = []
        obj.output['result2'] = obj.input * multiple


class TestMacroCommand(MacroCommand):
    NAME = 'CMD_TEST_MARCRO'

    def __init__(self):
        super().__init__()
        self.add_subcommand(TestSubCommand1)
        self.add_subcommand(TestSubCommand2)


class TestProxy(Proxy):
    NAME = 'PROXY_TEST'

    def __init__(self, name=None, data=None):
        super().__init__(name, data)


class TestMediator(Mediator):
    NAME = 'MEDIATOR_TEST'

    @property
    def notifications(self):
        return [ 'sth_interest' ]

    def handle_notification(self, note):
        if note.name == 'sth_interest':
            test_obj = note.body.get('obj')
            test_obj.output = test_obj.input + note.body.get('increase')


class ControllerTest(unittest.TestCase):
    def test_001_singleton(self):
        alpha = Controller()
        beta = Controller.get_instance()
        self.assertIs(alpha, beta)

class ModelTest(unittest.TestCase):
    def test_001_singleton(self):
        alpha = Model()
        beta = Model.get_instance()
        self.assertIs(alpha, beta)

class ViewTest(unittest.TestCase):
    def test_001_singleton(self):
        alpha = View()
        beta = View.get_instance()
        self.assertIs(alpha, beta)

class CommandTest(unittest.TestCase):
    def test_001_register_and_execute_and_remove_command(self):
        obj = TestObject(4)
        facade = Facade.get_instance()
        facade.register_command(TestSimpleCommand.NAME, TestSimpleCommand)
        self.assertTrue(facade.has_command(TestSimpleCommand.NAME))

        facade.send_notification(TestSimpleCommand.NAME, {
            'obj': obj,
            'increase': 8,
        })
        facade.remove_command(TestSimpleCommand.NAME)
        self.assertEqual(obj.output, 4 + 8)

    def test_002_register_and_remove_and_execute_command(self):
        obj = TestObject(1)
        facade = Facade.get_instance()
        # Test SimpleCommand
        facade.register_command(TestSimpleCommand.NAME, TestSimpleCommand)
        facade.remove_command(TestSimpleCommand.NAME)
        facade.send_notification(TestSimpleCommand.NAME, {
            'obj': obj,
            'increase': 8,
        })
        self.assertIsNone(obj.output)
        self.assertNotEqual(obj.output, 1 + 8)

        # Test MacroCommand
        obj = TestObject(2)
        facade.register_command(TestMacroCommand.NAME, TestMacroCommand)
        facade.send_notification(TestMacroCommand.NAME, {
            'obj': obj,
            'increase': 9,
            'multiple': 3,
        })
        self.assertEqual(obj.output.get('result1'), 2 + 9)
        self.assertEqual(obj.output.get('result2'), 2 * 3)


class ProxyTest(unittest.TestCase):
    def test_001_register_and_retrieve_and_remove_proxy(self):
        facade = Facade.get_instance()
        facade.register_proxy(TestProxy(data=TestObject(8)))
        self.assertTrue(facade.has_proxy(TestProxy.NAME))

        proxy = facade.retrieve_proxy(TestProxy.NAME)
        self.assertIsNotNone(proxy)
        self.assertIsNotNone(proxy.data)
        self.assertEqual(proxy.data.input, 8)
        self.assertEqual(proxy.data.output, None)

    def test_002_register_and_remove_and_retrieve_proxy(self):
        facade = Facade.get_instance()
        facade.register_proxy(TestProxy(data=TestObject(6)))
        self.assertTrue(facade.has_proxy(TestProxy.NAME))

        facade.remove_proxy(TestProxy.NAME)
        self.assertFalse(facade.has_proxy(TestProxy.NAME))
        self.assertIsNone(facade.retrieve_proxy(TestProxy.NAME))

    def test_003_raise_assert_not_none(self):
        self.assertRaises(AssertionError, Proxy, None, TestObject(0))
        self.assertRaises(AssertionError, Proxy, 'PROXY_TEST', None)
        self.assertRaises(AssertionError, Proxy, None, None)


class MediatorTest(unittest.TestCase):
    def test_001_register_and_retrieve_and_notify_and_remove_mediator(self):
        # Test register mediator
        facade = Facade.get_instance()
        facade.register_mediator(TestMediator(view_component=TestObject(8)))
        self.assertTrue(facade.has_mediator(TestMediator.NAME))

        # Test retrieve mediator
        mediator = facade.retrieve_mediator(TestMediator.NAME)
        self.assertIsNotNone(mediator)
        self.assertIsNotNone(mediator.view_component)
        self.assertEqual(mediator.view_component.input, 8)
        self.assertEqual(mediator.view_component.output, None)

        # Test notify interest
        test_obj = TestObject(6)
        facade.send_notification('sth_interest', {
            'obj': test_obj,
            'increase': 8,
        })
        self.assertEqual(test_obj.output, 6 + 8)

        # Test notify non-interest
        test_obj = TestObject(6)
        facade.send_notification('sth_not_interest', {
            'obj': test_obj,
            'increase': 8,
        })
        self.assertNotEqual(test_obj.output, 6 + 8)

        # Test remove mediator
        facade.remove_mediator(TestMediator.NAME)
        self.assertFalse(facade.has_mediator(TestMediator.NAME))
        test_obj = TestObject(6)
        facade.send_notification('sth_interest', {
            'obj': test_obj,
            'increase': 8,
        })
        # Test notify interest without register
        self.assertNotEqual(test_obj.output, 6 + 8)
        self.assertIsNone(test_obj.output)

    def test_002_raise_assert_not_none(self):
        self.assertRaises(AssertionError, Mediator, None, TestObject(0))
        self.assertRaises(AssertionError, Mediator, 'MEDIATOR_TEST', None)
        self.assertRaises(AssertionError, Mediator, None, None)


class FacadeTest(unittest.TestCase):
    def test_singleton(self):
        alpha = Facade()
        beta = Facade.get_instance()
        self.assertIs(alpha, beta)

    def test_001_command_and_mediator_and_proxy(self):
        facade = Facade.get_instance()
        # Register 
        facade.register_command(TestSimpleCommand.NAME, TestSimpleCommand)
        facade.register_command(TestMacroCommand.NAME, TestMacroCommand)
        facade.register_proxy(TestProxy(data=TestObject(3)))
        facade.register_mediator(TestMediator(view_component=TestObject(8)))

        self.assertTrue(facade.has_command(TestSimpleCommand.NAME))
        self.assertTrue(facade.has_command(TestMacroCommand.NAME))
        self.assertTrue(facade.has_proxy(TestProxy.NAME))
        self.assertTrue(facade.has_mediator(TestMediator.NAME))

        # retrieve
        proxy = facade.retrieve_proxy(TestProxy.NAME)
        mediator = facade.retrieve_mediator(TestMediator.NAME)
        self.assertIsNotNone(proxy)
        self.assertIsNotNone(mediator)

        # notify
        facade.send_notification(TestSimpleCommand.NAME, {
            'obj': proxy.data,
            'increase': 11,
        })
        self.assertEqual(proxy.data.output, 3 + 11)
        proxy.data.output = None
        facade.send_notification(TestMacroCommand.NAME, {
            'obj': proxy.data,
            'increase': 22,
            'multiple': 34,
        })
        self.assertEqual(proxy.data.output['result1'], 3 + 22)
        self.assertEqual(proxy.data.output['result2'], 3 * 34)

        facade.send_notification('sth_interest', {
            'obj': proxy.data,
            'increase': 23,
        })
        self.assertEqual(proxy.data.output, 3 + 23)
        del proxy, mediator

        # remove
        facade.remove_command(TestSimpleCommand.NAME)
        facade.remove_command(TestMacroCommand.NAME)
        facade.remove_mediator(TestMediator.NAME)
        facade.remove_proxy(TestProxy.NAME)
        self.assertFalse(facade.has_command(TestSimpleCommand.NAME))
        self.assertFalse(facade.has_command(TestMacroCommand.NAME))
        self.assertFalse(facade.has_mediator(TestMediator.NAME))
        self.assertFalse(facade.has_proxy(TestProxy.NAME))

        # notify but not response
        body = {
            'obj': TestObject(4),
            'increase': 11,
        }
        facade.send_notification(TestSimpleCommand.NAME, body)
        self.assertNotEqual(body['obj'].output, 4 + 11)

        body = {
            'obj': TestObject(8),
            'increase': 22,
            'multiple': 34,
        }

        facade.send_notification(TestMacroCommand.NAME, body)
        self.assertIsNone(body['obj'].output)

        body = {
            'obj': TestObject(6),
            'increase': 334,
        }
        facade.send_notification('sth_interest', body)
        self.assertNotEqual(body['obj'].output, 6 + 34)
        del body, facade



#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''
from logging import NOTSET, DEBUG, INFO, WARNING, ERROR, FATAL, CRITICAL
from logging import Handler
from unipark import requests
import six
import json
import traceback
import platform

COLORS = {
    CRITICAL:   'red',
    FATAL:      'red',
    ERROR:      'red',
    WARNING:    'orange',
    INFO:       'blue',
    DEBUG:      'green',
    NOTSET:     'grey',
}

RED_DOT =       b'\xf0\x9f\x94\xb4'.decode('utf-8')
ORANGE_DOT =    b'\xf0\x9f\x9f\xa0'.decode('utf-8')
BLUE_DOT =      b'\xf0\x9f\x94\xb5'.decode('utf-8')
GREEN_DOT =     b'\xf0\x9f\x9f\xa2'.decode('utf-8')
WHITE_DOT =     b'\xe2\x9a\xaa'.decode('utf-8')

EMOJIS = {
    CRITICAL:   RED_DOT,
    FATAL:      RED_DOT,
    ERROR:      RED_DOT,
    WARNING:    ORANGE_DOT,
    INFO:       BLUE_DOT,
    DEBUG:      GREEN_DOT,
    NOTSET:     WHITE_DOT,
}

class FeishuLogHandler(Handler):
    def __init__(self, \
                webhook_url, \
                project='unipark', \
                sender='UniPark', \
                fail_silent=True, \
                level=NOTSET):
        Handler.__init__(self, level)
        self.fail_silent = fail_silent
        self.webhook_url = webhook_url
        self.project = project.upper()
        self.sender = sender

    def build_trace(self, record):
        if record.exc_info[0]:
            return ''.join(traceback.format_exception(*record.exc_info))
        else:
            return ''.join(traceback.format_stack())

    def emit(self, record):
        message = six.text_type(self.format(record)) \
                if not record.exc_info \
                else six.text_type(record.msg)
        color = COLORS.get(record.levelno, 'blue')
        title = f'{ EMOJIS.get(record.levelno, BLUE_DOT) }[{ record.levelname.upper() }] ' \
                f'{ self.project } - { platform.node()}'
        content = f'Dear <at id=all>All</at>, a message from [{ self.project }] is received. Please pay attention to it!'
        card = {
            'msg_type': 'interactive',
            'card': {
                'config':  {
                    'wide_screen_mode': True,
                },
                'header': {
                    'template': color,
                    'title': {
                        'tag': 'plain_text',
                        'content': title,
                    },
                },
                'elements': [
                    {
                        'tag': 'div',
                        'text': {
                            'tag': 'lark_md',
                            'content': content,
                        },
                    },
                    {
                        'tag': 'hr',
                    },
                    {
                        'tag': 'div',
                        'fields': [
                            # Project
                            {
                                'is_short': True,
                                'text': {
                                    'tag': 'lark_md',
                                    'content': f'**Project:**\n{ self.project }',
                                },
                            },
                            # Computer
                            {
                                'is_short': True,
                                'text': {
                                    'tag': 'lark_md',
                                    'content': f'**Computer:**\n{ platform.node() }',
                                },
                            },
                            # Empty line
                            {
                                'is_short': False,
                                'text': {
                                    'tag': 'lark_md',
                                    'content': ''
                                },
                            },
                            {
                                'is_short': False,
                                'text': {
                                    'tag': 'lark_md',
                                    'content': f'**Message:** ({ record.pathname }:{ record.lineno })'
                                },
                            } if not record.exc_info else {
                                'is_short': False,
                                'text': {
                                    'tag': 'lark_md',
                                    'content': f'**Stacktrace:** ({ record.pathname }:{ record.lineno })',
                                },
                            },
                            {
                                # Message
                                'is_short': False,
                                'text': {
                                    'tag': 'lark_md',
                                    'content': f'{ message }',
                                },
                            } if not record.exc_info else {
                                # Exception
                                'is_short': False,
                                'text': {
                                    'tag': 'plain_text',
                                    'content': f'{ self.build_trace(record) }',
                                },
                            },
                        ],
                    },
                    {
                        'tag': 'hr',
                    },
                    {
                        'tag': 'note',
                        'elements': [
                            {
                                'tag': 'lark_md',
                                'content': f'From: { self.sender }',
                            },
                        ],
                    },
                ],
            },
        }
        try:
            resp = requests.post(self.webhook_url,
                    headers={'Content-Type': 'application/json; charset=utf-8'},
                    json=card)
            assert resp.status_code == 200, \
                    f'Call Feishu API failed.'
            assert resp.json().get('code') == 0, \
                    f'Feishu API Error: { resp.json().get("msg") }'
        except Exception as e:
            if self.fail_silent:
                pass
            else:
                raise e



#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''

__all__ = [
    'gettext',
    'ngettext',
    'pgettext',
    'npgettext',
    'lgettext',
    'lngettext',
]
import os

class Translation(object):
    def __init__(self, *args, **kwargs):
        self.setup()

    def __getattr__(self, name):
        return getattr(self._lang, name)

    def setup(self, domain='messages', localedir='.', languages=None, \
            *args, **kwargs):
        import gettext
        assert os.path.isdir(localedir), \
                f'Directory {localedir} is not exists.'
        assert gettext.find(domain, localedir, languages) is not None, \
                f'Translation files for locale is not found.'
        self._lang = gettext.translation(domain, localedir, languages)

_trans = Translation()
del Translation


def gettext(message):
    return _trans.gettext(message)

def ngettext(singular, plural, n):
    return _trans.ngettext(singular, plural, n)

def pgettext(context, message):
    return _trans.pgettext(context, message)

def npgettext(context, singular, plural, n):
    return _trans.npgettext(context, singular, plural, n)

def lgettext(message):
    return _trans.lgettext(message)

def lngettext(singular, plural, n):
    return _trans.lngettext(singular, plural, n)



#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''
from .facade import Facade
from .observer import Notifier


class Mediator(Notifier):
    NAME = None

    def __init__(self, name=None, view_component=None, *args, **kwargs):
        self._facade = Facade.get_instance()
        self._name = name or self.NAME
        self._view_component = view_component
        assert self._name is not None, \
                f'Mediator name cannot be "None"'
        assert self._view_component is not None, \
                f'Mediator view_component cannot be "None"'

    @property
    def facade(self):
        return self._facade

    @property
    def name(self):
        return self._name

    def send_notification(self, note_name, note_body=None, note_type=None):
        self.facade.send_notification(note_name, note_body, note_type)

    @property
    def view_component(self):
        return self._view_component

    @view_component.setter
    def view_component(self, value):
        self._view_component = value

    @property
    def notifications(self):
        return []

    def handle_notification(self, note):
        pass

    def on_register(self):
        pass

    def on_remove(self):
        pass



#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''
from .facade import Facade
from .observer import Notifier

class Command(Notifier):
    def __init__(self, *args, **kwargs):
        self._facade = Facade.get_instance()

    @property
    def facade(self):
        return self._facade

    def execute(self, note):
        raise NotImplementedError(self)


class SimpleCommand(Command):
    def execute(self, note):
        pass

class MacroCommand(Command):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._subcommands = []

    def add_subcommand(self, command_class):
        self._subcommands.append(command_class)

    def execute(self, note):
        for command_class in self._subcommands:
            command_class().execute(note)



#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''
import unittest

from unipark.cli import cli
import click

@click.command('hello')
@click.option('--name', prompt='Your name')
def hello(name):
    click.echo(f'Hello {name}!')

if __name__ == '__main__':
    hello()

class CliTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_001_assert_is_command(self):
        self.assertIsInstance(cli, click.core.Command)

    def test_002_add_subcommand(self):
        cli.add_command(hello)
        self.assertTrue(True)



#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''
__all__ = (
    'Controller',
    'Model',
    'View',
    'Notification',
)
from .observer import Observer, Notification
class Controller(object):
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, 'instance'):
            cls.instance = super(Controller, cls).__new__(cls, *args, **kwargs)
        return cls.instance

    @staticmethod
    def get_instance():
        return Controller()

    def __init__(self, *args, **kwargs):
        if not hasattr(self, '_view') \
                or self._view is None:
            self._view = View.get_instance()
        if not hasattr(self, '_commands') \
                or self._commands is None:
            self._commands = {}

    @property
    def view(self):
        return self._view

    @property
    def commands(self):
        return self._commands

    def register_command(self, note_name, command_class):
        if note_name not in self.commands:
            self.view.register_observer(note_name, \
                    Observer(self.execute_command, self))
        self.commands[note_name] = command_class

    def execute_command(self, note):
        command_class = self.commands.get(note.name)
        if command_class:
            command_instance = command_class()
            command_instance.execute(note)

    def remove_command(self, note_name):
        if self.has_command(note_name):
            self.view.remove_observer(note_name, self)
            del self.commands[note_name]

    def has_command(self, note_name):
        return note_name in self.commands


class Model(object):
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, 'instance'):
            cls.instance = super(Model, cls).__new__(cls, *args, **kwargs)
        return cls.instance

    @staticmethod
    def get_instance():
        return Model()

    def __init__(self, *args, **kwargs):
        if not hasattr(self, '_proxies') \
                or self._proxies is None:
            self._proxies = {}

    @property
    def proxies(self):
        return self._proxies

    def register_proxy(self, proxy):
        self.proxies[ proxy.name ] = proxy
        proxy.on_register()

    def retrieve_proxy(self, proxy_name):
        return self.proxies.get(proxy_name)

    def remove_proxy(self, proxy_name):
        proxy = self.proxies.pop(proxy_name, None)
        if proxy:
            proxy.on_remove()
        return proxy

    def has_proxy(self, proxy_name):
        return proxy_name in self.proxies


class View(object):
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, 'instance'):
            cls.instance = super(View, cls).__new__(cls, *args, **kwargs)
        return cls.instance

    @staticmethod
    def get_instance():
        return View()

    def __init__(self, *args, **kwargs):
        if not hasattr(self, '_observers') \
                or self._observers is None:
            self._observers = {}
        if not hasattr(self, '_mediators') \
                or self._mediators is None:
            self._mediators = {}

    @property
    def observers(self):
        return self._observers

    @property
    def mediators(self):
        return self._mediators

    def register_observer(self, note_name, observer):
        if not note_name in self.observers:
            self.observers[note_name] = []
        self.observers[note_name].append(observer)

    def notify_observers(self, notification):
        observers = self.observers.get(notification.name, [])[:]
        for observer in observers:
            observer.notify_observer(notification)

    def remove_observer(self, note_name, notify_context):
        observers = self.observers.get(note_name)
        for i in range(len(observers) - 1, -1, -1):
            if observers[i].compare_notify_context(notify_context):
                observers.pop(i)

        if len(observers) == 0:
            self.observers.pop(note_name)

    def register_mediator(self, mediator):
        if mediator.name in self.mediators:
            return

        self.mediators[mediator.name] = mediator
        if len(mediator.notifications) > 0:
            observer = Observer(mediator.handle_notification, mediator)

            for note_name in mediator.notifications:
                self.register_observer(note_name, observer)
        mediator.on_register

    def retrieve_mediator(self, mediator_name):
        return self.mediators.get(mediator_name)

    def remove_mediator(self, mediator_name):
        mediator = self.mediators.get(mediator_name)
        for note_name in tuple(self.observers.keys()):
            observers = self.observers[note_name]
            for i in range(len(observers) - 1, -1, -1):
                if observers[i].compare_notify_context(mediator):
                    observers.pop(i)

            if len(observers) == 0:
                self.observers.pop(note_name)

        if mediator is not None:
            self.mediators.pop(mediator_name)
            mediator.on_remove()

        return mediator

    def has_mediator(self, mediator_name):
        return mediator_name in self.mediators




#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''

class Notification(object):
    def __init__(self, name, body=None, type=None):
        self._name = name
        self._body = body
        self._type = type

    @property
    def name(self):
        return self._name

    @property
    def body(self):
        return self._body

    @body.setter
    def body(self, value):
        self._body = value

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, value):
        self._type = value

    def __repr__(self):
        return self.name


class Notifier(object):
    def send_notification(self, note_name, note_body=None, note_type=None):
        pass

    @property
    def name(self):
        return self._name if hasattr(self, '_name') else None


class Observer(object):
    def __init__(self, notify_method, notify_context):
        self._method = None
        self._context = None

        self.notify_method = notify_method
        self.notify_context = notify_context

    @property
    def notify_method(self):
        return self._method

    @notify_method.setter
    def notify_method(self, value):
        self._method = value

    @property
    def notify_context(self):
        return self._context

    @notify_context.setter
    def notify_context(self, value):
        self._context = value

    def notify_observer(self, note):
        self.notify_method(note)

    def compare_notify_context(self, obj):
        return obj is self.notify_context




#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''
__all__ = [ 'cli', ]

import click
import os

@click.group(invoke_without_command=True)
@click.option('-d', '--debug/--no-debug', \
        is_flag=True, default=False, show_default=True)
@click.pass_context
def cli(ctx, debug):
    settings = ctx.obj.get('SETTINGS', None)

    if settings is not None \
            and debug != settings('DEBUG', default=False):
        os.environ.update({ 'DEBUG': str(debug) })

if __name__ == '__main__':
    cli(obj={})



#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''
from setuptools import setup, find_packages
import os

LONG_DESCRIPTION = None
with open('README.md', 'r') as fp:
    LONG_DESCRIPTION = fp.read()

REQUIREMENTS = None
with open('requirements/runtime.txt', 'r') as fp:
    REQUIREMENTS = [ line.strip() for line in fp.readlines() \
            if not line.strip().startswith('#') ]

VERSION = None
with open('VERSION', 'r') as fp:
    VERSION = fp.read().strip()

with open('unipark/version.py', 'w') as fp:
    fp.write(f'# Generated by "setup.py" file.\n'
            f'VERSION=\'{ VERSION }\'')

def find_package_data(data_dir, exts=('*',)):
    paths = []
    for root, dirs, files in os.walk(data_dir):
        for file in files:
            if not exts \
                    or '*' in exts \
                    or [ ext for ext in exts if file.endswith(ext) ]:
                paths.append(os.path.join(root, file))
    return paths

if __name__ == '__main__':
    setup(
        name='unipark',
        version=VERSION,
        description='Common library of UniPark',
        long_description=LONG_DESCRIPTION,
        long_description_content_type='text/markdown',
        author='Park Lam',
        author_email='lqmonline@gmail.com',
        url='https://www.unipark.io',
        license='Apache2.0',
        install_requires=REQUIREMENTS,
        packages=find_packages(
            include=[ 'unipark*', ],
            exclude=[ 'ez_setup', 'test*', ]),
        package_data={
            'unipark': find_package_data('data') \
                    + [ 'licenses.txt', ]},
        entry_points={
            'console_scripts': [ 'unipark = unipark:main' ],
        },
    )



#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''
from . import Model, View, Controller
from .observer import Notification, Notifier

class Facade(Notifier):
    def __new__(cls, *args, **kwargs):
        if not hasattr(cls, 'instance'):
            cls.instance = super(Facade, cls).__new__(cls, *args, **kwargs)
        return cls.instance

    @staticmethod
    def get_instance():
        return Facade()

    def __init__(self, *args, **kwargs):
        if not hasattr(self, '_controller') \
                or self._controller is None:
            self._controller = Controller.get_instance()
        if not hasattr(self, '_model') \
                or self._model is None:
            self._model = Model.get_instance()
        if not hasattr(self, '_view') \
                or self._view is None:
            self._view = View.get_instance()

    @property
    def controller(self):
        return self._controller

    @property
    def model(self):
        return self._model

    @property
    def view(self):
        return self._view

    def register_command(self, note_name, cmd_cls):
        self.controller.register_command(note_name, cmd_cls)

    def remove_command(self, note_name):
        self.controller.remove_command(note_name)

    def has_command(self, note_name):
        return self.controller.has_command(note_name)

    def register_proxy(self, proxy):
        self.model.register_proxy(proxy)

    def retrieve_proxy(self, name):
        return self.model.retrieve_proxy(name)

    def remove_proxy(self, name):
        return self.model.remove_proxy(name)

    def has_proxy(self, name):
        return self.model.has_proxy(name)

    def register_mediator(self, mediator):
        self.view.register_mediator(mediator)

    def retrieve_mediator(self, name):
        return self.view.retrieve_mediator(name)

    def remove_mediator(self, name):
        return self.view.remove_mediator(name)

    def has_mediator(self, name):
        return self.view.has_mediator(name)

    def send_notification(self, note_name, note_body=None, note_type=None):
        self.notify_observers(
            Notification(
                note_name, note_body, note_type
            )
        )

    def notify_observers(self, note):
        self.view.notify_observers(note)



#!/usr/bin/env python3
# coding: utf-8
'''
Author: Park Lam <lqmonline@gmail.com>
'''
__all__ = [ '__version__', 'main' ]

import os
import click
from .cli import cli
from .version import VERSION as __version__
from .utils import generate_password, generate_salt

_BASE_DIR = os.path.dirname(os.path.abspath(__file__))

@cli.command('version')
@click.pass_context
def version(ctx):
    click.echo(f'{ __version__ }')

@cli.command('password')
@click.option('-l', '--length', default=32)
@click.pass_context
def password(ctx, length):
    click.echo(generate_password(length=length))

@cli.command('salt')
@click.pass_context
def salt(ctx):
    click.echo(generate_salt())

def main():
    from .config import settings
    cli(obj={ 'SETTINGS': settings })

if __name__ == '__main__':
    main()
